;; 9 February 2005 - added the :print-object keyword to the defstruct
;; of ckpt to allow a clean print of the checkpoints. Before it wasn't
;; possible to print checkpoints because the print of the first
;; element was causing a recursive call to itself when it was printing
;; the next of its prev (that is itself). So the printing process
;; wasn't terminating.

;;       Last change    August 1993
(in-package lib)

;                     Checkpoint and Retraction
;
; In order to handle checkpointing and retracting of propositions entered
; into the system, changes must be stored. They are stored at an
; atomic level, as there is no way to predict how far a change may
; propagate from a higher level.

;;; The ckpt structure defines an entry in the checkpoint stack, one for 
;;; each user defined or anonymous checkpoint.

(defstruct (ckpt (:print-object
		  (lambda (o s)
		    (format s "CKPT: ~a STACK-PTR: ~a PREV: ~a~%"
			    (ckpt-name o)
			    (ckpt-stack-ptr o)
			    (ckpt-prev o)))))
  (stack-ptr nil)	;pointer into the rewind stack, to the element
			; in the stack which was added just after this 
			; checkpoint was made (so this can be nil).

  name			;checkpoint name (even anon checkpoints have names)
  anon-p                ;t if this is an anonymous ckpt, nil otherwise
  (next nil)		;next (as in added after this one) checkpoint 
  (prev nil))		;previous checkpoint

; a rewind-item contains information on how to get to the changed item,
; and its previous value.

(defstruct rewind-item
  (what nil)			; what was changed (field item) (aref a index)
				;  (get 'item 'property) (gethash key hasharray)
  (oldvalue nil)		; value it had before the change
  )

(defvar *num-anon* 0)	;Number of anonymous checkpoints currently used
(defvar *max-anon* 0)	;Maximum number of anonymous checkpoints to maintain
(defvar *oldest-anon* nil)	;pointer to oldest anonymous checkpoint

(defvar *ckpts* nil) ;points to the list of checkpoints currently set. The
			 ; first point is the most recent.
(defvar *rewind-stack* nil) ;;; Rewind stack, where each entry is 
                         ; (undo-function . undo-function-args)

;;; checkpointing-p is true if we are currently checkpointing.
(defmacro checkpointing-p () '(and *ckpts* t))

;;; Get-last-checkpoint returns the name of the last checkpoint set. 
;;; It's possible
;;; to see if a checkpoint has been set between two invocations of this
;;; function, by #'eq comparing the values returned.
(defmacro get-last-checkpoint () `(ckpt-name *ckpts*))
       

; save-change retrieves the old value for what, and saves what (how to get it)
; and the old value in a rewind item. This is added at the beginning of the
; rewind list, as all changes have to be done in reverse.

(defun save-change (what &aux
			 (oldvalue (eval what))
			 (new-rewind (make-rewind-item :what what
						       :oldvalue oldvalue)))
  (start-checkpoint 'rewind new-rewind)
  )

; change saves a rewind-item for the change requested, and then does the change

(defun change (what newvalue)
  (when (checkpointing-p) (save-change what))
  (eval (list 'setf what (list 'quote newvalue)))
  )

; undo replaces the item specified by the rewind-item ritem with its old value

(defun rewind (ritem &aux
		   (what (rewind-item-what ritem))
		   (oldvalue (rewind-item-oldvalue ritem)))
  (eval (list 'setf what  (list 'quote oldvalue)))
  )

; change-array saves the old value in index of array, and puts newvalue there

(defun change-array (array index newvalue)
  (change (list 'aref array index) newvalue)
  )

; change-hash saves the old value for item in hasharray, and puts newvalue there

(defun change-hash (item hasharray newvalue)
  (change (list 'gethash (list 'quote item) hasharray) newvalue)
  )

; change-field saves the old value in fieldname for item, puts newvalue there

(defun change-field (fieldname item newvalue)
  (change (list fieldname item) newvalue)
  )

; change-var saves the old value in a variable, sets it to newvalue 

(defun change-var (item newvalue)
  (change (eval (list 'quote item)) newvalue)
  )

; change-property

(defun change-property (item property value)
  (change (list 'get (list 'quote item) (list 'quote property)) value)
  )

; increment

(defun increment (fieldname item &aux (value (eval (list fieldname item))))
  (change (list fieldname item) (+ value 1)))

;;; Checkpointing is done using two stacks, one to maintain the list of
;;; checkpoints, and the other a rewind stack to record the things that 
;;; have to be undone when retracting.
  


;;; start-checkpoint puts an item on the rewind stack.
;;; When retracting, undo-function is applied to undo-args.

(defun start-checkpoint (undo-function &rest undo-args)
  (when *ckpts*
	(push (cons undo-function undo-args) *rewind-stack*)

	;; Use this entry for checkpoints taken just previous
	;; to it.

	(do ((ckpt *ckpts* (ckpt-prev ckpt)))
	    ((or (not ckpt) (ckpt-stack-ptr ckpt)))
	    (setf (ckpt-stack-ptr ckpt) *rewind-stack*)))
  ) 
  

;;; anon-checkpoint sets an anonymous checkpoint, removing the
;;; oldest such checkpoint if the number of anonymous checkpoints 
;;; exceeds the maximum.
 
(defun anon-checkpoint ()
  (when (> *max-anon* 0)
	(add-checkpoint (gensym "ANON") :anon t)
	(if (not *oldest-anon*) (setq *oldest-anon* *ckpts*))
	(if (> (incf *num-anon*) *max-anon*) (remove-anon)))
  nil)


;;; Remove-anon removes the oldest anonymous checkpoint from the
;;; checkpoint stack.

(defun remove-anon ()
  (when *oldest-anon*
	(if (ckpt-prev *oldest-anon*)
	    (setf (ckpt-next (ckpt-prev *oldest-anon*)) 
		  (ckpt-next *oldest-anon*)))

	(if (ckpt-next *oldest-anon*)
	    (setf (ckpt-prev (ckpt-next *oldest-anon*))
		  (ckpt-prev *oldest-anon*))
	    (setq *ckpts* (ckpt-prev *oldest-anon*)))

	(setq *oldest-anon* (ckpt-next *oldest-anon*))
	(do nil
	    ((or (not *oldest-anon*) (ckpt-anon-p *oldest-anon*)))
	    (setq *oldest-anon* (ckpt-next *oldest-anon*)))

	;; If there are no checkpoints before the next oldest anonymous
        ;; checkpoint then we can discard all the undo items put on the
        ;; rewind stack before that checkpoint.

	(if (and *oldest-anon* (not (ckpt-prev *oldest-anon*)))
	    (if (ckpt-stack-ptr *oldest-anon*) 
	        (setf (cdr (ckpt-stack-ptr *oldest-anon*)) nil)
		(setq *rewind-stack* nil)))

	(decf *num-anon*))
  )


;;; Add-checkpoint adds a checkpoint to the checkpoint stack, setting
;;; up the prev and next links as required.

(defun add-checkpoint (name &key (anon nil) &aux new-ckpt)
  (setq new-ckpt (make-ckpt :name name
			    :anon-p anon
			    :prev *ckpts*))
  (if *ckpts* (setf (ckpt-next *ckpts*) new-ckpt))
  (setf *ckpts* new-ckpt)
  name)


;;; With an integer argument, enable anonymous checkpointing with
;;; *max-anon* set to that number.  With no argument, or a string
;;; argument, set up a named checkpoint and return the name.  The name
;;; is generated by gensym, and includes the string if given.

(defun checkpoint (&optional arg)
  (cond ((or (not arg) (stringp arg))
	 (add-checkpoint (if arg
			     (gensym arg)
			   (gensym))))
	((and (integerp arg) (>= arg 0))
	 (do nil ((<= *num-anon* arg)) (remove-anon))
	 (setq *max-anon* arg)
	 (if (not *ckpts*) (setq *rewind-stack* nil))
	 nil)
	(t (print-line "The argument to CHECKPOINT, if given, should be "
		       "either a string or a positive integer.")
	   nil)))

(defun no-checkpointing ()
  (setq *oldest-anon* (setq *ckpts* (setq *rewind-stack* nil)))
  (setq *num-anon* (setq *max-anon* 0))
  nil)	



;;; With an integer argument, rewind past that many anonymous
;;; checkpoints.  With a symbol argument, rewind to the checkpoint
;;; with that name (a checkpoint's name is the symbol returned by
;;; (checkpoint) when the checkpoint was set).  If the keyword
;;; argument :remove is non-nil, remove the checkpoint from the stack,
;;; else leave it active.
	
(defun retract (arg &key remove)
  (cond ((and (integerp arg) (>= arg 0)) (retract-anon arg))
	((symbolp arg) (retract-named remove arg))
	(t  (print-line "The argument to RETRACT should be either a "
			"symbol or a positive integer.")))
  nil)


;;; Retract-anon retracts up to `num' anonymous checkpoints.

(defun retract-anon (num)
  (if (= *num-anon* 0)
      (print-line "There are no anonymous checkpoints to retract.")
      (do ((num-done 0))
	  ((or (= num-done num) (= *num-anon* 0)) 
	   (if (< num-done num) 
	       (print-line "There were only " num-done
			   " anonymous checkpoints to retract.")))
	
	(rewind-rewind-stack *ckpts*)

	(if (ckpt-anon-p *ckpts*)
	    (progn
	      (incf num-done)
	      (decf *num-anon*)
	      (if (eq *oldest-anon* *ckpts*) (setq *oldest-anon* nil)))
	  (print-line "Retracting past the named checkpoint "
		      (ckpt-name *ckpts*)))
	
	(if (setq *ckpts* (ckpt-prev *ckpts*))
	    (setf (ckpt-next *ckpts*) nil))))
  )


;;; Retract-named retracts up to a named checkpoint.
;;;
;;;	if remove-name is true, the checkpoint itself is removed
;;;	sym is the name of the checkpoint to retract to

(defun retract-named (remove-name sym &aux ckpt)
  (when (setq ckpt (do ((cp *ckpts* (ckpt-prev cp)))
		       ((not cp) (print-line sym " is not the name of a checkpoint."))
		     (if (eq (ckpt-name cp) sym) (return cp))))
    (do nil
	(nil)
      (rewind-rewind-stack *ckpts*)
      (if (eq *ckpts* ckpt) (return))
      
      (if (ckpt-anon-p *ckpts*)
	  (progn
	    (decf *num-anon*)
	    (if (eq *oldest-anon* *ckpts*) (setq *oldest-anon* nil)))
	(print-line "Retracting past the named checkpoint "
		    (ckpt-name *ckpts*)))
      
      (if (setq *ckpts* (ckpt-prev *ckpts*))
	  (setf (ckpt-next *ckpts*) nil)))
    
    (if remove-name
	(if (setq *ckpts* (ckpt-prev *ckpts*))
	    (setf (ckpt-next *ckpts*) nil)))))


;;; Rewind-stack undoes items on the rewind stack.
;;;
;;;	ckpt is a checkpoint which indicates the place in the
;;;		rewind stack to rewind to.

(defun rewind-rewind-stack (ckpt)
  (do nil
      ((not (ckpt-stack-ptr ckpt)))
    (apply (caar *rewind-stack*) (cdar *rewind-stack*))
    
    ;; If we've rewound to the requested checkpoint, make sure we 
    ;; disable all identical checkpoints i.e., which pointed to
    ;; the same entry in the rewind stack.
    
    (do ((cp ckpt (ckpt-prev cp)))
	((or (not cp) (not (eq (ckpt-stack-ptr cp) *rewind-stack*))))
      (setf (ckpt-stack-ptr cp) nil))
    
    (pop *rewind-stack*))
  )


;; These functions for the question checkpoint would more logically go
;; in ../story/qa.lisp, but then the lib package would need to use
;; epilog package, which would cause a circularity.

(defvar *question-ckpt* nil "name of the current question checkpoint,
or nil if there isn't one.  Don't mess with this variable except via
checkpoint-question and rewind-question.")

(defun checkpoint-question ()
  (when *question-ckpt*
      (print-line "Warning: multiple question checkpoints being set.  " 
		  "All but the latest are ignored."))
  (setq *question-ckpt* (checkpoint "QUESTION")))

(defun rewind-question ()
  (when *question-ckpt*
    (retract *question-ckpt* :remove t)
    (setq *question-ckpt* nil)))

(defun earliest-checkpoint (ckpts)
  (if ckpts
      (let ((prev (ckpt-prev ckpts)))
	(if prev
	    (earliest-checkpoint prev)
	  ckpts))
    nil))

(defun rewind-all ()
  (let ((c (earliest-checkpoint *ckpts*)))
    (if c (retract (ckpt-name c)))))   ;should this have :remove t?


;;; Just for debugging

(defun display-ckpts (&optional (msg  " ") (rewind-stack nil))
  (print-line "Checkpoint state " msg)
  (when rewind-stack (print-line "Rewind-stack:")
	(mapc #'(lambda (item) (print-line "    " (car item) " " (cadr item)))
	      *rewind-stack*) )
  (print-line "Checkpoints:") 
  (do ((ckpt *ckpts* (ckpt-prev ckpt)))
      ((not ckpt))
      (print-line (if (ckpt-name ckpt) (ckpt-name ckpt) "anon" ) " " 
		  (if (ckpt-stack-ptr ckpt)
		      (caar (ckpt-stack-ptr ckpt))
		      "no stack pointer")
		  (if (eq ckpt *oldest-anon*)
		      " oldest anonymous checkpoint"
		      "")))
  )

(set-display-function 'checkpoint 'display-ckpts "Displays active checkpoints")
 
