;       Last change    August 1993
(in-package response)

; Verbalize - the stage that actually produces the English. Here it is divided
; into several phases:
;   1) create fragments for all literals in clauses and combine
;      according to type of clause
;      combine all clause fragments - the result is a set of sentence
;      fragments.
;         (files: fragment, combine, trans, verbalize)
;   2) rearrange fragments for negations and passivize if necessary
;         (files: fill, fragment, verbalize)
;   3) for each sentence fragment, fill in any gaps - determiners, 
;      noun phrase info (may have to look up in the topic skeleton), verbs, etc
;         (files: fill)
;   4) go through filled fragments and determine correct lexical entries for
;      each item - result is a list of words.
;         (files: print, lex)
;   5) a postprocessor goes through this list, changing "a" to "an"
;      before a vowel, and combining words like "every" "thing" to "everything"
;         (files: print)
;   6) The sentence is printed.
;         (files: print)


(defvar *sorted-constants* nil)
(defvar *verbalize-wffs* nil )
(defvar *answer* nil "Contains the current answer")
(defvar *already-universal* nil "Notes what number the last variable got")
(defvar *already-neg* nil "Notes whether a negation has been used")
(defvar *there-is* nil)
(defvar *nomed-verbs* nil)
(defvar *mod-frags* nil)
 
; verbalize takes the given list of clauses and generates English from them

(defun verbalize (answer logforms)
  ; returns list of sentences. The first consists of the answer and
  ; the first logical form translation; the others, the logical form
  ; translations of other elements in logforms
  (reset-indent "" 0)
  (setq *said* nil)
  (setq *functions* nil)
  (setq *verbalize-wffs* nil)
  (setq *answer* answer)
  (if (car logforms)
      (sentence-trans logforms) 
    (print-answer answer)
    )
  nil)

(defun sentence-trans (wff-list)
  (setq *fragment-number* 0)
  (setq *fragment-list* nil)
  (setq *entity-list* nil)
  (setq *entity-info* nil)
  (setq *conditional* nil)
  (single-sentence (car wff-list) t)
  ; additional sentences
  (mapcar 'single-sentence (cdr wff-list))
  )

(defun single-sentence (wff-list &optional first &aux res winf fst)
  (when (traced-p 'verbalization)
        (print-line "  Response: Logical formulae to verbalize:")
	(mapcar #'(lambda (c) 
			  (print-line "  " c ": " (getp c 'PRINT)))
		wff-list))
  (setq *verbalize-wffs* (append *verbalize-wffs* wff-list))
  (setq *entity-list* (remove-if 'epi-variable *entity-list*))
  (setq *entity-info* 
	(remove-if #'(lambda (a) (epi-variable (car a))) *entity-info*))
  (setq *already-universal* nil)
  (setq *there-is* nil)
  (setq *already-neg* nil)
  (setq *first* t)
  (setq res (mapcar 'wff-fragment wff-list))
  ; make the fragments into sentences to say, if they won't combine
  ; now lookup anything that needs looking up
  (do ((v (car (set-difference *entity-list* (mapcar 'car *entity-info*)))
          (car (set-difference *entity-list* (mapcar 'car *entity-info*)))))
      ((null v))
      ; default info if nothing found ...
      (if (epi-variable v) (setq winf nil) (setq winf (lookup-info v)))
      (update-entity-info v winf)
      (when winf (setq res (append res (if winf (list winf)))))
      )
  (setq res (cond ((apply 'conjunct-frags res)) (t res)))
  ;; should this be done before or after lookup? 
  (setq res (mapcar #'(lambda (r)
			(setq *there-is* nil)
			(if (equal (getp r 'frag-type) 'S) r (make-into-S r))) res))  
  ;; print the sentence now, including answer if necessary
  (setq fst t)
  (dolist (r res) 
    (setq *nomed-verbs* nil)
    (setq *mod-frags* nil)
    (setq r (fill-gaps r))
    (when (traced-p 'verbalization)
      (print-line "  Sentence fragment to say: ")
      (print-frag r 2))
    (if (and first fst)
	(progn
	  (print-it-first r)
	  (setq fst nil))
      (print-and r))
    (dolist (n *nomed-verbs*) (setp n 'aspect nil))
    (dolist (f *mod-frags*) 
      (dolist (p (getp f 'save-slist)) (setp f (car p) (cadr p)))
      (remprop f 'save-slist))
    )
  )

; lookup-info finds info to use for a particular entity - if it is a anme,
; the name is used, otherwise info is looked up in the knowledge base

(defun lookup-info (item &aux sort class wtrans tr)
  (when (traced-p 'retrieval)
	(print-line "  Response: Looking up info for " item))
  (when (setq tr (if (symbolp item) (getp item 'trans) nil))
	(return-from lookup-info
		     (make-frag 'NP (cons 'name item) (cons 'proper-noun tr))))
  (cond ((equal (setq sort (getp item 'sort)) 'episode)
         (setq class (list item 'episode-specialist)))
        ((equal sort 'propos)
         (setq class (list item '_ 'object)))
        (t (setq class (list item 'type-specialist))))
  (dolist (w (get-class class))
          (unless (or (member w *verbalize-wffs*) (filter-check w t))
		  (when (traced-p 'retrieval) 
			(print-line "  Response: Retrieved " w ": " (print-info w)))
		  (when (and (type-lit-p w)
			     (not (equal (find-sort item) 'episode))
			     (equal (getp w 'wff-type) 'episodic))
			(setq w (car (getp w 'arguments))))
                  (setq wtrans (wff-fragment w))
                  (return))
          )
  wtrans)

; How to verbalize sentences

; frag is np with pp below, pp is the pp frag, usepp is the of pp (may be same
; as pp, or first part of it), np is the np of usepp

(defun has-part (frag pp usepp np &aux new parts)
  (cond ((setq parts (fragment-parts pp))
	 (setp pp 'parts (cdr parts))
	 (if (equal (list-length (fragment-parts pp)) 1)
	     (setq pp (car (fragment-parts pp))))
	 )
	(t (setq pp nil)))
  (setp frag 'PP pp)
  (set-subject nil frag)
  (set-subject t np)
  (setq *already-universal* 'singular)
  (setq new
  (make-frag 'S 
	     (cons 'NP np)
	     (cons 'VP (make-frag 'VP (cons 'name (fragment-name np) )
				  (cons 'negated (or (fragment-negated frag) 
						     (fragment-negated usepp)
						     (fragment-negated np)))
				  (cons 'verb 'have)
				  (list 'objects frag)))
	     ))
  (setp frag 'negated nil)
  (setp usepp 'negated nil)
  (setp np 'negated nil)
  new)

(defun make-into-NP (frag &optional subject &aux ftype nfrag)
  (cond ((or (not (symbolp frag)) (null (setq ftype (getp frag 'frag-type))))
	 (setq nfrag (make-frag 'NP (cons 'name frag) (cons 'subject subject)))
	 nfrag)
	(t (case (fragment-fragtype frag)
		 (NP frag)
		 ((PP AP) 
		  (make-frag 'NP (cons 'name (fragment-name frag))
		    (cons 'head (fragment-head frag))
		    (cons 'subject (or (fragment-subject frag) subject))
		    (cons 'number (fragment-number frag))
		    (cons ftype frag)))
		 (S (nominalize frag subject))
		 (otherwise frag))
	   ))
  )

(defun nominalize (frag subject &aux vp h)
  (cond ((setq vp (fragment-VP frag))
	 ; make into pres part
	 (unless (and (setq h (fragment-head frag))
		      (equal (fragment-fragtype h) 'advp))
	 (setp vp 'aspect 'prespart)
	 (push vp *nomed-verbs*) )
	 (setp frag 'advp nil)
	 (setp frag 'advs nil)
	 (make-frag 'NP (cons 'name (fragment-name frag))
		    (cons 'head (fragment-head frag))
		    (cons 'subject (or (fragment-subject frag) subject))
		    (cons 'S frag)))
	(t (make-frag 'NP (cons 'name (fragment-name frag))
		    (cons 'head (fragment-head frag))
		    (cons 'subject (or (fragment-subject frag) subject))
		    ))
	)
  )


(defun make-into-S (frag &aux s ftype np vp)
  (if (or (not (symbolp frag)) (null (setq ftype (getp frag 'frag-type))))
      (return-from make-into-S (make-frag 'S (cons 'name frag))))

  (case ftype
	(S (when (and (setq np (fragment-np frag))
		      (null (fragment-VP frag))
		      (setq vp (fragment-relc np)))
		 (setp frag 'VP vp)
		 (setp np 'RELC nil)
		 (setp vp 'frag-type 'VP)
		 (setp vp 'head frag)
		 )
	   (when (and (fragment-ADVP frag)
		      (not (member frag *said* :test 'same-name))
		      (not (and (setq s (find-entity frag))
				(equal (fragment-fragtype s) 'S)
				(getp s 'VP)))
		      (epi-variable (fragment-name frag))
		      (not (or (fragment-parts frag) (fragment-VP frag)))
		      (not (fragment-VP (fragment-S (fragment-ADVP frag)))))
		 (setq frag (S-advp-into-S-pp frag)))
	 frag)
	((ADVP ADVS) (make-frag 'S (cons 'name (fragment-name frag))
			 (cons 'subject (fragment-subject frag))
			 (cons ftype frag)) )
	((NP PP AP) (interesting-S frag))
	(VP (make-frag 'S (cons 'name 'removed)
		       (cons 'NP (make-into-NP (fragment-name frag) t))
		       (cons 'VP frag)))
	(otherwise frag)
	)
  )

(defun interesting-S (frag &aux vp nfrag tr pp np parts)
  ; if NP and S below NP - this may be one of those ( )_ p1 things
  (cond ((and (equal (fragment-fragtype frag) 'NP) (fragment-S frag))
	 (if (or (fragment-AP frag) (fragment-noun frag) (fragment-PP frag))
	     (let ((np (fragment-s frag))
		   vp)
		  (setp frag 'S nil)
		  (setq vp (make-into-VP frag))
		  (make-frag 'S (cons 'NP np) (cons 'VP vp))
		  )
	     (let ()
		  (setf (fragment-head (fragment-S frag)) (fragment-head frag))
		  (fragment-S frag))
	     ))

	((and (equal (fragment-fragtype frag) 'NP)
	      (member (getp (fragment-name frag) 'term-type)
		      '(record quoted-expression function
			       quasi-quoted-expression))
	      (fragment-print frag))
	 (setp frag 'frag-type 'S)
	 frag)

	((and (equal (fragment-fragtype frag) 'NP)
	      (null (noun-info frag))
	      (null (fragment-AP frag))
	      (null (fragment-PP frag))
	      (null (Fragment-Relc frag)))
	 (setp frag 'frag-type 'S)
	 frag)

	((and (equal (fragment-fragtype frag) 'NP)
	      (fragment-Relc frag))
	 (setq vp (fragment-Relc frag))
	 (setp frag 'Relc nil)
	 (setp vp 'frag-type 'VP)
	 (make-frag 'S (cons 'NP frag) (cons 'VP vp))
	 )

	((and (equal (fragment-fragtype frag) 'NP)
	      (fragment-noun frag)
	      (not (equal (fragment-objtype frag) 'constant))
	      (or (fragment-AP frag) (fragment-PP frag)))
	 (cond ((fragment-PP frag)
		(setq vp (make-into-VP (fragment-PP frag)))
		(setp frag 'PP nil)
		(make-frag 'S (cons 'NP frag) (cons 'VP vp)))
	      (t (setq vp (make-into-VP (fragment-AP frag)))
		(setp frag 'AP nil)
		(make-frag 'S (cons 'NP frag) (cons 'VP vp)))
	       ))

	((and (symbolp (fragment-name frag))
	      (setq tr (getp (fragment-name frag) 'trans)))
	 (set-subject nil frag)
	 (make-frag 'S (cons 'head (fragment-head frag))
		    (cons 'NP (make-frag 'NP
					 (cons 'name (fragment-name frag)) 
					 (cons 'proper-noun tr)
					 (cons 'subject t)))
		    (cons 'VP (make-frag 'VP
					 (cons 'name (fragment-name frag)) 
					 (cons 'verb 'be)
					 (list 'objects frag)))) )

	; NP with PP ... of x becomes x has ...
	((and (equal (fragment-fragtype frag) 'NP)
	      (setq pp (fragment-PP frag))
	      (equal (fragment-prep pp) 'of)
	      (setq np (fragment-NP pp))
	      (not (check-noun-property (fragment-noun frag) 'group))
	      (not (check-noun-property (fragment-noun frag) 'member))
	      (not (member (fragment-noun frag) '(part cause start end)))
	      (null (fragment-pre pp))
	      (null (fragment-adv pp))
	      )
	 (has-part frag pp pp np))

	((and (equal (fragment-fragtype frag) 'NP)
	      (setq pp (fragment-PP frag))
	      (not (member (fragment-prep pp) '(of for)))
	      (setq np (fragment-NP pp)))
	 (setp frag 'PP nil)
	 (make-frag 'S (cons 'NP frag) (cons 'VP (make-into-VP pp))))

	; more than one PP - use first as part of NP, rest as objects of VP
	((and (or (and (equal (fragment-fragtype frag) 'NP)
		       (setq pp (fragment-PP frag))
		       (setq nfrag frag))
		  (and (equal (fragment-fragtype frag) 'PP)
		       (setq pp frag)
		       (setq nfrag 
			     (make-frag 'NP (cons 'name (fragment-name frag)))))
		  )
	      (setq parts (fragment-parts pp)))
	 (let ((ps (cdr parts)))
	      (setp nfrag 'PP (car parts))
	      (setp nfrag 'subject t)
	      (make-frag 'S (cons 'head (fragment-head frag))
			 (cons 'NP nfrag)
			 (cons 'VP (make-frag 'VP 
					      (cons 'name (fragment-name frag))
					      (cons 'verb 'be)
					      (cons 'objects ps)))
			 )))

	((and (equal (fragment-fragtype frag) 'NP)
	      (fragment-name frag)
	      (not (equal (fragment-name frag) 'removed))
	      (not (or (epi-variable (fragment-name frag))
		       (and (getp (fragment-name frag) 'indef)
			    (not (getp (fragment-name frag) 'unique))))))
	 (make-frag 'S 
		    (cons 'NP (make-into-NP (fragment-name frag) t))
		    (cons 'VP (make-into-VP frag))))

	; There is ...
	((and nil (not *there-is*)
	      (equal (fragment-fragtype frag) 'NP)
	      (or (null (fragment-name frag))
		  (and (epi-variable (fragment-name frag))
		       (equal (fragment-quantifier frag) 'E))
		  (if (fragment-parts frag)
		  (eval (append (list 'and) (mapcar #'(lambda (a) 
			      (or (skolem-constant-p (fragment-name a))
				  (and (epi-variable (fragment-name a))
				       (equal (fragment-quantifier a) 'E))
				  )) (fragment-parts frag)))))) )
	 (setp frag 'required t)
	 (setq *there-is* t)
	 (set-subject nil frag)
	 (make-frag 'S (cons 'NP (make-frag 'NP (cons 'name 'there)
					    (cons 'print "there")
					    (cons 'subject t)))
		    (cons 'VP (make-frag 'VP (cons 'name (fragment-name frag))
					 (cons 'verb 'be)
					 (list 'objects frag)))
		    ))

	; no verb ??
	((equal (fragment-name frag) 'removed)
	 (make-frag 'S (cons 'NP (make-frag 'NP 
					    (cons 'name (fragment-name 'frag))
					    (cons 'subject t)))
                        
		    (cons 'VP (make-frag 'VP 
					 (cons 'name (fragment-name 'frag))
					 (cons 'verb 'removed)
					 (list 'objects frag)))))
	; if conjucntion, make into sentence conjunction
	((and (fragment-parts frag) 
	      (not (eval (append (list 'and)
			    (mapcar #'(lambda (a) (same-name frag a)) 
				    (fragment-parts frag))))))
	 (setp frag 'parts (mapcar 'make-into-S (fragment-parts frag)))
	 (setp frag 'frag-type 'S)
	 frag)

	; default case - make name into NP, fragment into object of VP
	(t (make-frag 'S 
			 (cons 'NP (make-into-NP (fragment-name frag) t))
			 (cons 'VP (make-into-VP frag)))
	   ))
  )

(defun lambda-frag (frag &aux name) 
  (cond ((null frag) t) 
	((and (epi-variable (setq name (fragment-name frag))) 
	      (getp name 'lambda))   
	 t) 
	(t nil)) 
  ) 

(defun make-into-VP (frag &aux name ftype)
  (if (or (not (symbolp frag)) (null (setq ftype (getp frag 'frag-type))))
      (return-from make-into-VP (make-frag 'VP (cons name frag)
					   (cons 'verb 'be)
					   (cons 'objects (list frag)))))

  (case ftype
	(VP frag)
	(S (fragment-VP frag))
	((PP NP AP)
	 (let ((vpfrag (make-frag 'VP (cons 'name (fragment-name frag))
				 (cons 'head (fragment-head frag))
				 (cons 'negated (fragment-negated frag))
				 (cons 'verb 'be)
				 (list 'objects frag))))
	      (set-subject nil frag) 
	      (setp frag 'negated nil)
	      vpfrag)) 
	(otherwise frag)
	))

(defun VP-into-RELC (frag &aux newfrag)
  (setq newfrag
  (case (fragment-fragtype frag)
	(VP (setp frag 'frag-type 'RELC)
	    (make-frag 'NP (cons 'name (fragment-name frag)) 
		    (cons 'head (fragment-head frag))
		    (cons 'Relc frag)))
	(S 
	 (if (fragment-VP frag)
	     (let ((nfrag (fragment-NP frag))
		   (vfrag (fragment-VP frag)))
		  (setp vfrag 'frag-type 'RELC)
		  (if (fragment-negated frag) 
		      (setp vfrag 'negated t))
		  (setp nfrag 'RelC vfrag)
		  (setp frag 'VP nil)
		  (setp vfrag 'head nfrag)
		  nfrag)
	     frag))
	(otherwise frag)
	))
  (when (member (fragment-head newfrag) (fragment-objects newfrag))
      (print-line "here it is!!!! how did it happen?")
      (read)
      (setp newfrag 'objects nil))
  newfrag)

;  Special section for conditional statements

(defun test-only (pre post)
  (and pre (equal (fragment-fragtype (car post)) 'NP)
       ;(not (fragment-parts (car post)))
       (not (equal (find-sort (fragment-name (car post))) 'episode))
       (member (fragment-fragtype (car pre)) '(S VP))
       (or (same-name (fragment-np (car pre)) (car post))
	   (same-name (car (fragment-objects (fragment-VP (car pre))))
		      (car post)))
       (not (equal (fragment-verb 
		    (if (and (equal (fragment-fragtype (car pre)) 'S) 
			     (fragment-VP (car pre)))
			(fragment-VP (car pre)) 
			(car pre))) 
		   'be))
       )
  )

(defun combinable (a b &aux
			 (subja (fragment-name a))
			 (subjb (get-subject b))
			 (obj1b (get-first-object b)))
  ; something common between a and b
  (and subja subjb
  (or (equal (fragment-name a) (fragment-name b))
      (equal subja subjb) 
      (equal subja obj1b)
      ))
  )

(defun if-then-test (pre post)
  (and pre post
       (or (equal (fragment-name (car pre)) 'removed)
	   (and (or (member (fragment-fragtype (car pre)) 
			    '(S VP))
		    (and (equal (fragment-fragtype (car pre)) 'NP)
			 (or (equal (find-sort (fragment-name (car pre)))
				'episode)
			     (member (fragment-noun (car pre))
				     '(episode event)))))
		(member (fragment-fragtype (car post)) '(S VP))
		(if (epi-variable (get-subject (car pre)))
		    (not (equal (get-subject (car pre))
				(get-subject (car post))))
		    t)
		(not (equal (fragment-name (car pre))
			    (fragment-name (car post)))) 
		))
       )
  )

(defun cond-statement (pre post)
  (cond ;((apply 'conjunct-frags (append pre post)))
	((and pre 
	      (if (equal (fragment-name (car pre)) 
			       (fragment-name (car post)))
		  (not (or (fragment-advp (car pre))
			   (fragment-advp (car post))))
		  t)
	     (member-if 
	      #'(lambda (f) (member (fragment-fragtype f) '(VP S))) post))
	(setq pre (mapcar 'VP-into-RELC pre) )
	(cond ((apply 'conjunct-frags (append pre post)))
	      (t (append pre post)))
	 )
        ((apply 'conjunct-frags (append pre post)))
	(t (append pre post))
	 )
  )

(defun massage-conditional (pre post adverb &aux resp ppas res)
  ; handle "only type do x" - otherwise make post into S
  (if (test-only pre post)
      (let ()
	   (if (fragment-negated (car post))
	       (let ()
		    (add-number 'plural (car post))
		    (setq *already-neg* t)
		    (setq *already-universal* 'plural)
		    (setq *first* nil)
		    (add-determiner "no" (car post))
		    (setp (car post) 'negated nil))
	     (setp (car post) 'adv (combine-l "only" (fragment-adv (car post))))
	     )
	   (add-number 'plural (car post) t)
	   (setq *already-universal* 'plural)
	   ; passivize if necessary - that type must be the subject
	   (when (not (or (equal (fragment-name (car post)) 
			     (get-subject (car pre)))
			  (equal (fragment-name (car post))
				 (fragment-name (car pre)))))

	       (setq pre (if (setq ppas (do-passive (car pre)))
			     (append (list ppas) (cdr pre))
			     pre))
		 )

	   (setq post (append (list (car post))
			      (mapcar 'make-into-VP (cdr post))))
	   (setq res (apply 'conjunct-frags (append post pre)))
	   (unless (or (null res) (equal res (append post pre)))
		   (setq pre res)
		   (setq post nil))
	   )
      (unless (and (equal (fragment-fragtype (car post)) 'S)
		   (equal (fragment-fragtype (car pre)) 'S)
		   (fragment-VP (car pre))
		   (member-if #'(lambda (a) (same-name (car post) a)) pre))
	      (setq post (mapcar 'make-into-S post))) ;VP
      )

  ; if pre is NP, AP or PP, add quantinfo
  ; else add adverb to post
  (when adverb
	(setp (car post) 'adv (combine-l adverb (fragment-adv (car post)))))

  ;Passivize pre or post if that will make them combine better
  (when (and pre post
	     (setq resp (try-passive (car pre) (car post))))
	(setq pre (cons (car resp) (cdr pre)))
	(setq post (cons (cadr resp) (cdr post)))
	)
  (cons pre post)
  )

; choose-post-subj looks through all post conditions and picks one with
; verb information on it as the subject - otherwise the first post condition

(defun choose-post-subj (post &aux
			     subj)
  (dolist (p post)
	  (when (fragment-VP p)
	        (setq subj p)
		(return subj)))
  (if subj
      (append (list subj) (remove subj post))
      post)
  )

; choose-pre-subj looks through all pre conditions and picks one with
; verb information on it as the subject - otherwise the first pre condition

(defun choose-pre-subj (pre postsubj &aux
			    subj)
  (dolist (p pre)
	  (when (equal (fragment-NAME p) 
		       (fragment-NAME postsubj))
		(setq subj p)
		(return p)))
  (if subj
      (append (list subj) (remove subj pre))
      pre)
  )

; if-then creates a sentence of the form "if pre then post" and is used to
; print implications when other methods fail (esp if they are too complicated)

(defun if-then (pre post &aux npor usenp)
  (setq *already-universal* 'singular)
  ; if pre then post
  (setq pre (cond ((apply 'conjunct-frags pre)) (t pre)))
  (add-number 'singular (car pre) t)
  (add-number 'singular (car post))
  (setq *already-universal* 'singular)

  ;build if part  - if pre parts don't fit together - NPand
  (if (cdr pre)
      (setq usenp (make-conj nil 'and pre))
      (setq usenp (car pre))
      )

  ; now for then part - use NPor
  (if (cdr post)
      (setq npor (make-conj nil 'and post))
      (setq npor (car post)))

  (make-frag 'S (cons 'name 'removed)
	     (cons 'pre 'if)
	     (cons 'parts (mapcar 'make-into-S (list usenp npor) ))
	      )
   )


