;       Last change    August 1993
(in-package time)
;                  DEFINITIONS

; The time net consists of time-nodes (one for each time point), which are
; arranged in a hash table for easy access. They point to each other via
; time-links. The time-links contain points back to the points themselves
; so that when renumbering takes place, each psuedo time need only change
; in one place. They also contain duration and strictness information.
; 
; Each chain has a meta-node associated with it, which contains a list of
; time-links for cross chain links from that chain.
; Meta-nodes are arranged in an array by chain number.
;
; There are also event-nodes (one for each event) which contain the names
; of the start and end points of the event. Names are used so that if a
; point is equated with another and replaced by it, there is no need to 
; check all events to see if they also use it.
; Event-nodes are also arranged in a hash table for easy access (by event name)


;  Eg.  t1 <= t2,  t3 < t2

;    ______
;    |meta|
;    |node|--
;    | 1  | |
;    ______ |first
;           |
;           v
;	 ______        ______        ______
;	 |time|  desc  |time|  anc  |time|
;	 |node|<------>|link|<------>|node|
;	 | t1 |        | <= |        | t2 |
;	 ______        ______        ______
;                                        ^
;                                        |
;    ______                              |
;    |meta|                              |
;    |node|--                   xanc    |
;    | 2  |  |              -------------
;    ______  |             |
;     f|     |             |
;     i|     |connections  |
;     r|      -------      v
;     s ----         |   ______
;     t	    |         -->|time|
;	    |       ____>|link|
;	    |      |x    | <  |
;           v      |d    ______
;	  ______   |e
;   	  |time|   |s
;	  |node|<-- c
;	  | t3 |
;	  ______


(defstruct time-node
  (name nil)			; name of time point
  (chain nil)			; meta-node for chain that point belongs to
  (psuedo nil)			; psuedo time
  (min-point-name nil)		; name of earliest point equal to
  (max-point-name t)		; name of latest point equal to
  (absolute-min nil)		; absolute time minimum
  (absolute-max nil)		; absolute time maximum
  ;; Warning: the ancestor lists and alternate-names list must not be
  ;; modified destructively, because their structure is shared when
  ;; nodes are copied into a new environment.  See copy-chain-nodes.
  (ancestors nil)		; in chain ascendants
  (xancestors nil)		; cross chain ascendants
  (descendants nil)		; in chain descendants
  (xdescendants nil)		; cross chain descendants
  (alternate-names nil)		; list of names of points collapsed into this
  (environment *environment*)   ; environment to which it belongs
  )

;   
;      ______         ______        ______
;      |time|   min   |time|  max   |time|
;      |node|<--------|node|------->|node|
;      ______         ______        ______
;                    / | | \
;           xanc   /  | |  \ xdesc
;           -------- --  --  --------
;          |         |a    |d        |
;          |         |n    |e        |
;          |         |s    |s        |
;          v         vc    vc        v
;       ______   ______   ______   ______
;       |time|   |time|   |time|   |time|
;       |link|   |link|   |link|   |link|
;       ______   ______   ______   ______
;         |        |        |        |
;         |        |        |        |
;         v        v        v        v
;       ______   ______   ______   ______
;       |time|   |time|   |time|   |time|
;       |link|   |link|   |link|   |link|
;       ______   ______   ______   ______
;


(defstruct time-link
  (from-point-name nil)		; name of point link is from
  (to-point-name nil)		; name of point link is to
  (strict nil)			; indicates strictness (t or nil)
  (duration-min nil)		; minimum duration between
  (duration-max nil)		; maximum duration between
  )

;
;     ______            ______            ______
;     |time|   from     |time|   to       |time|
;     |node|<-----------|link|----------->|node|
;     ______            ______            ______
;

(defstruct meta-node
  (chain-number nil)		; chain number
  (first nil)			; first time point of chain
  (connections nil)		; cross chain links
  (environment *environment*)   ; env to which it belongs
  )

;; number of chains created.  A chain number should be unique among
;; all environments (if the same number appears in two environments,
;; then the one in the more specific environment shadows the one in
;; the less specific one), so this variable is not shadowed in
;; simulations.

(defvar *chain-count* 0)        

; Newchain creates a new chain for the next available chain number and
; updates the meta graph with it.  The chain structure is returned.

(defun newchain ()
  (let ((node (change-hash *chain-count*
			   (time-net-meta-graph (writeable-net))
			   (make-meta-node :chain-number *chain-count*))))
    (change-var '*chain-count* (1+ *chain-count*))
    node))

;
;   ______            ______
;   |meta|   first    |time|
;   |node|----------->|node|
;   ______            ______
;     | connections
;     --------
;             |
;             v
;           ______           ______
;           |time|           |time|
;           |link|---------->|link|
;           ______           ______
;


(defstruct time-net
  (identifier nil)		; net identifier (context)
  (time-graph nil)		; hash table of time points
  (meta-graph nil)		; hash table of meta nodes
  (event-table nil)		; hash table containing event nodes
  (environment *environment*)
  )

;
;    ______                 ______              ______
;    |time|  time-graph     |hash|              |time|
;    |net |---------------->|tbl |------------->|node|     ______
;    ______                 |    |              ______     |time|
;     |    \                |    |------------------------>|node|
;    e|     \               ______                         ______
;    v|      --------
;    e|  meta-graph |
;    n|             v
;    t|           _______          ______
;    -|           |hash |          |meta|
;    t|           |tbl  |--------->|node|     ______
;    a|           |     |          ______     |meta|
;    b|           |     |-------------------->|node|
;    l|           _______                     ______
;    e|
;     v
;     ______              _______
;     |hash|              |event|
;     |tbl |------------->|node |        _______
;     |    |              ______         |event|
;     |    |---------------------------->|node |
;     ______                             _______
;

(defstruct event-node
  (name nil)			; name of event 
  (start nil)			; name of start point
  (end nil)			; name of end point
  (environment *environment*)
  )

(defconstant newpsuedo '1 "Psuedo time for first element on chain")
(defconstant psuedo-increment '1000 "Increment between nodes")
