;       Last change    August 1993
(in-package time)
;                 EVALUATION Routines
;
; The evaluation routines work by comparing the given relation against the
; most strict one that holds between the two points. For requests dealing
; with more than two points, the request is split into several requests
; involving two points each.


; evaluate-absolute determines the tests required to determine if the
; given relation holds between the given points - at least one must be
; an absolute time
; determine-split is used to get the relations needed between the points
; (only the first two are required - four are returned in some cases)
; the tests are then done, and if any returns with nil or unknown,
; that is the result, otherwise true

(defun evaluate-absolute (a1 stem s1 s2 a2 a3 effort &aux
			      (predlist (determine-split stem s1 s2))
                              (preds (mapcar #'(lambda (a) 
				     (build-pred (car a) (cadr a) nil))
					   predlist))
			      (usepreds (subseq preds '0 '2))
                              (pts1 (abs-start a1))
	                      (pte1 (abs-end a1)))
  (if (equal stem 'between)
      (let ((pt2 (abs-end a2))
	    (pt3 (abs-start a3))
	    res1 res2)
	   (setq res1 (test-point-answer (car usepreds)
                                  (find-absolute pts1 pt2 effort)))	
	   (if (and (or (not (equal res1 'unknown))
			(traced-p 'epilog::time-search))
		    (traced-p 'epilog::time-point-eval))
	       (print-line  "    Time Specialist: Evaluated " 
				(print-info pts1) " "  (car usepreds) 
			  " " (print-info pt2) " to " res1))
	   (setq res2 (test-point-answer (cadr usepreds)
				  (find-absolute pte1 pt3 effort)))
	   (if (and (or (not (equal res2 'unknown))
			(traced-p 'epilog::time-search))
		    (traced-p 'epilog::time-point-eval))
	       (print-line  "    Time Specialist: Evaluated " 
				(print-info pte1) " " (cadr usepreds) 
			  " " (print-info pt3) " to " res2))
	   (combine-test-results res1 res2)
	   )

      ; for other stems: before after during contains overlaps etc

      (let ((pts2 (abs-start a2))
	    (pte2 (abs-end a2))
	    res1 res2)
	   (setq res1 (test-point-answer (car usepreds)
			             (find-absolute pts1 pts2 effort)))
	   (if (and (or (not (equal res1 'unknown))
			(traced-p 'epilog::time-search))
		    (traced-p 'epilog::time-point-eval))
	       (print-line "    Time Specialist: Evaluated " 
				(print-info pts1) " " (car usepreds) 
			  " " (print-info pts2) " to " res1))
	   (setq res2 (test-point-answer (cadr usepreds)
			             (find-absolute pte1 pte2 effort)))
	   (if (and (or (not (equal res2 'unknown))
			(traced-p 'epilog::time-search))
		    (traced-p 'epilog::time-point-eval))
	       (print-line "    Time Specialist: Evaluated " 
				(print-info pte1) " " (car usepreds) 
			  " " (print-info pte2) " to " res2))
	   (setq res1 (combine-test-results res1 res2))
	   ; should also check during a3
	   (if (or (null a3)
		   (not (equal res1 'yes)))
	       res1
	       (combine-test-results res1 
		   (evaluate-absolute a1 'during nil nil a3 nil effort))
	       )
	   )
      )
  )

; determine-tests returns a list of tests to be done to see if the given
; relation holds between the given points or events
; it uses determine split which returns the relations which must hold between
;     the start points
;     the end points
;     start of first and end of second
;     start of second and end of first
; determine-tests places the appropriate arguments in the tests with these
; predicates

(defun determine-tests (a1 stem s1 s2 a2 a3 &aux
			    (predlist (determine-split stem s1 s2))
                            (preds (mapcar #'(lambda (a) 
				     (build-pred (car a) (cadr a) nil))
				   predlist)))
  (if (member stem '(between))
      (let* ((a1start (get-start a1))
	     (a1end (get-end a1))
	     (a2end (get-end a2))
	     (a3start (get-start a3))
	     (endtest (build-pred 'before (combine-strict s1 s2)))
	     (s1e2 (car preds))
	     (e1s3 (cadr preds))
	     (todolist (list (list a1start s1e2 a2end))) )
	    (when (event a1)
	          (setq todolist (append todolist 
			               (list (list a1end e1s3 a3start)))))
	    (setq todolist (append todolist 
				   (list (list a2end endtest a3start))))
	    todolist)
      (let* ((a1start (get-start a1))
	     (a1end (get-end a1))
	     (a2start (get-start a2))
	     (a2end (get-end a2))
	     (s1s2 (car preds))
	     (e1e2 (cadr preds))
	     (s1e2 (caddr preds))
	     (e1s2 (cadddr preds))
	     (todolist (list (list a1start s1s2 a2start))))
	    (when (or (event a1)
		      (event a2)
		      (member stem '(during overlaps contains overlapped-by)))
	          (setq todolist (append todolist
			               (list (list a1end e1e2 a2end))
			               (list (list a1start s1e2 a2end))
			               (list (list a1end e1s2 a2start))))
		  )
	    (unless (null a3)
		    (setq todolist (append todolist 
					   (determine-tests a1 'during nil 
							    nil a3 nil))))
	    todolist)
      )
  )


; canttell returns t if there is no way to tell anything about the arguments
; because not enough of them are in the time graph.
; If even 2 out of the three are there, it is possible to at least get a
; negative answer (although no positive one will be possible).

(defun canttell (a1 a2 a3)
  (<= 2 (count-if-not #'null (mapcar #'(lambda (a) (missing a))
				(if (null a3) 
				    (list a1 a2)
				    (list a1 a2 a3)))))
  )

; evaluate-point-relations calls determine-tests to determine what tests
; need to be done to decide whether the given relation holds
; The tests are then done - if any comes back nil or unknown, that is the answer; otherwise it is true

(defun evaluate-point-relations (a1 stem s1 s2 a2 a3 effort)

  ; see if enough arguments in net to make a test worthwhile

  (if (canttell a1 a2 a3)
      'unknown

  ; if there are, determine the tests needed to answer the request, and
  ; do the tests

  (let*  ((todo (determine-tests a1 stem s1 s2 a2 a3))
          res majorres)

  ; all the relations given in todo must evaluate to true for a true result

  (dolist (item todo)
	  (let ((pred (cadr item))
	        (pt1 (car item))
		(pt2 (caddr item)))
	       (setq res (test-point-answer pred 
				       (find-point pt1 pt2 effort)))

	       (if (and (or (not (equal res 'unknown))
			    (traced-p 'epilog::time-search))
			(traced-p 'epilog::time-point-eval))
		   (print-line "    Time Specialist: Evaluated " 
				    (print-info pt1) " " pred 
				    " " (print-info pt2) " to " 
				    res))

	       (if (equal res 'no)
		   (return res)
		   (if (equal res 'unknown)
		       (setq majorres 'unknown))
		   )
	       )
	  )
   ; if we had to save an unknown result earlier, hoping for a more definite
   ; result, make sure that it comes out if needed

   (if (equal res 'no)
       'no
       (if (equal majorres 'unknown)
           'unknown
           res)
       )
  ))
  )

(defun get-num-value (arg &aux fn res)
  (setq fn (make-fn-arg 'epilog::value-of arg))
  (setq res (spec-eval-fn fn))
  (cond ((not (equal res fn)) (list res res))
	(t (let* ((minfn (make-fn-arg 'epilog::min-of arg))
		  (minres (spec-eval-fn minfn))
		  (maxfn (make-fn-arg 'epilog::max-of arg))
		  (maxres (spec-eval-fn maxfn)))
		 (if (equal minres minfn) (setq minres nil))
		 (if (equal maxres maxfn) (setq maxres nil))
		 (if (or minres maxres) (list minres maxres)) )))
  )

; test-minimum-duration determines if the minimum duration given in dur is
; at least testvalue. If it is, it returns t, if it cannot be because 
; the maximum is less than it, nil is returned. Otherwise it is unknown.

(defun test-minimum-duration (dur test &aux
				  (testvalue (cond ((symbolp test)
						    (get-num-value test))
						   ((listp test) test)
						   (t (list test test))))
				  (testmin (if testvalue (car testvalue)))
				  (testmax (if testvalue (cadr testvalue)))
				  (dmin (car dur))
				  (dmax (cadr dur)))
  (cond ((or (null dmin)
	     (equal dmin 'epilog::-infinity))
	 'unknown)
	((and testmax 
	     (>= dmin testmax))
	 'yes)
	((or (null dmax)
	     (equal dmax 'epilog::+infinity))
	 'unknown)
	((or (equal dmin 'epilog::+infinity)
	     (equal dmax 'epilog::-infinity)
	     (and testmin
		  (< dmin testmin))
	     (and testmin 
		  (< dmax testmin)))
	 'no)
	(t 'unknown)
	)
  )

; test-maximum-duration determines if the maximum duration given in dur is
; at most testvalue. If it is, it returns t, if it cannot be because 
; the maximum is greater than it, nil is returned. Otherwise it is unknown.

(defun test-maximum-duration (dur test &aux
				  (testvalue (cond ((symbolp test)
						    (get-num-value test))
						   ((listp test) test)
						   (t (list test test))))
				  (testmin (if (atom testvalue)
					       testvalue
					       (car testvalue)))
				  (testmax (if (atom testvalue)
					       testvalue
					       (cadr testvalue)))
				  (dmin (car dur))
				  (dmax (cadr dur)))
  (cond ((or (null dmax)
	     (equal dmax 'epilog::+infinity))
	 'unknown)
	((and testmin
	      (<= dmax testmin))
	 'yes)
	((or (null dmin)
	     (equal dmin 'epilog::-infinity))
	 'unknown)
	((or (equal dmin 'epilog::+infinity)
	     (equal dmax 'epilog::-infinity)
	     (and testmax
		  (> dmin testmax))
	     (and testmax
		  (> dmax testmax)))
	 'no)
	(t 'unknown)
	)
  )

; evaluate-durations tries to determine if the given relation actually
; holds between a1 and a2 with duration a3
; First it determines the durtion between them. If the duration comes back
; without giving any information, it checks to see if the opposite relation
; holds between the points (eg. at-least-before checks for after). If
; so, the answer is nil.

(defun evaluate-durations (a1 reln a2 a3 effort)

  (if (or (missing a1)
	  (missing a2))
      'unknown

  ; determine duration in appropriate direction

  (let* ((testdur a3)
	 dur)

  (case reln
	((at-least-before at-most-before exactly-before)
         (setq dur (calc-duration (get-end a1) 
				  (get-start a2) effort)))
	((at-least-after at-most-after exactly-after)
         (setq dur (calc-duration (get-end a2) 
				  (get-start a1) effort)))
	)

  ; if inconclusive result, check to see if opposite relation holds.

  (if (or (null dur)
	  (and (equal (car dur) 'epilog::-infinity)
	       (equal (cadr dur) 'epilog::+infinity)))
      (case reln
	    ((at-least-before at-most-before exactly-before)
	     (if (equal (evaluate-point-relations a1 'before '1 nil a2 
						 nil effort) 'no)
		 'no
                 'unknown)
	     )
	    ((at-least-after at-most-after exactly-after)
	     (if (equal (evaluate-point-relations a1 'after '1 nil 
						 a2 nil effort) 'no)
		 'no
                 'unknown)
	     )
	    )

      ; otherwise look at the duration retuned.

      (case reln
	    ((at-least-before at-least-after)
	     (test-minimum-duration dur testdur)
	     )
	    ((at-most-before at-most-after)
	     (test-maximum-duration dur testdur)
	     )
	    ((exactly-before exactly-after)
	     (combine-test-results (test-minimum-duration dur testdur)
	                           (test-maximum-duration dur testdur))
	     )
	    )
      )
  ))
  )

; evaluate tries to evaluate whether or not the given temporal predicate 
; applied to arglist is true, false or unknown in the subnet associated with
; netname
; It divides the propositons into three classes:
;    those with absolute times -> evaluate-absolute
;    those indicating duration -> evaluate-durations
;    others                    -> evluate-point-relations

(defun evaluate (netname lit negp pred arglist effort &aux
		         (reln (find-symbol (symbol-name pred) 'time))
		         (a2 (cadr arglist)))
  (declare (ignore lit))
  (assert (eq *current-netname* netname))
  ;; I'm worried by the fact that there used to be two ways for the
  ;; netname information to get into the specialists, via the netname
  ;; argument and via *current-netame*.  I want to be sure that they're
  ;; consistent. -ANK

  (if (or (null arglist)
	  (non-constant-arg arglist)
	  (not (member (find-sort a2) '(epilog::time epilog::episode)))
	  )
      'unknown
      (eval-with-neg negp reln arglist effort)
      )
  )

; eval-with-neg evaluates the given literal, taking into account whether it
; is negated. If so, the positive literal is evaluated, and the result reversed.

(defun eval-with-neg (negp reln args effort &aux
			  (posans (evaluate-time reln args effort)))
  (cond ((equal posans 'yes)
	 (if negp
	     'no
	     'yes))
	((equal posans 'no)
	 (if negp	
	     'yes
	     'no))
	(t posans))
  )



; evaluate-time does the actual evaluation of a positive literal. No further
; checking needs to be done to see if this is in the temporal specialist
; domain.

(defun evaluate-time (reln arglist effort &aux
		         (plist (split-time-pred reln))
		         (stem (stem-of plist))
		         (s1 (first-strict plist))
		         (s2 (second-strict plist))
		         (a1 (car arglist))
		         (a2 (cadr arglist))
		         (a3 (caddr arglist))
			 (result 'unknown))

; (defun time::evaluate-time (reln arglist effort &aux
;         plist stem s1 s2 a1 a2 a3 result)
;   (print-line "    Time Specialist: Evaluate-time " reln)
;	(setq plist (split-time-pred reln))  (print-line plist)
;	(setq stem (stem-of plist))     (print-line stem)
;	(setq s1 (first-strict plist))  (print-line s1)
;	(setq s2 (second-strict plist)) (print-line s2)
;	(setq a1 (car arglist))         (print-line a1)
;	(setq a2 (cadr arglist))        (print-line a2)
;	(setq a3 (caddr arglist))       (print-line a3)
;	(setq result 'unknown)          (print-line result)

  (unless (null arglist)

  (when (traced-p 'epilog::time-eval)
        (if (null a3)
            (print-line "    Time Specialist: Evaluating " 
			   (print-info a1) " " reln " " (print-info a2))
            (print-line "    Time Specialist: Evaluating " 
			   (print-info a1) " " reln " " 
			   (print-info a2) " " (print-info a3) )
	    )
	(if (not (equal *current-netname* '/)) 
	    (print-it "    in subnet " *current-netname*)) 
      )

  ; if any absolute time arguments, use evaluate-absolute

  (if (or (absolute-time a1)
	  (absolute-time a2)
	  (and (equal stem 'between)
	       (absolute-time a3)))
      (setq result (evaluate-absolute a1 stem s1 s2 a2 a3 effort))

      ; if duration predicate, use evaluate-durations

      (if (member reln '(at-least-before at-least-after
                         at-most-before at-most-after
			 exactly-before exactly-after))
	  (if (number-arg a3) 
	      (setq result (evaluate-durations a1 reln a2 a3 effort))
	      (setq result 'unknown))

	  ; otherwise do point comparisons only

          (setq result (evaluate-point-relations a1 stem s1 s2 a2 a3 effort))
	  )
      )
  )
  (if (and (or (not (equal result 'unknown))
	       (traced-p 'epilog::time-search))
	   (traced-p 'epilog::time-eval))
      (if (null a3)
          (print-line "    Time Specialist: Evaluated " (print-info a1 )
			   " " reln " " (print-info a2) " to " result )
          (print-line "    Time Specialist: Evaluated " (print-info a1)
			   " " reln " " (print-info a2) " " (print-info a3)
			   " to " result)
	  )
      )
  result
  )
