;       Last change    August 1993
(in-package time)
;                  PSUEDO-TIMES

; Within a chain, each time node has a pseudo-time associated with it that
; gives the position of that point with respect to the other ones in the
; chain. This number is arbitrarily assigned when the point is added to
; maintain the relative positions (higher number after, lower before).
; When adding between two points, if there isn't enough space left, the
; psuedo times can be renumbered without affecting the relative position
; of the points.
;
; To further complicate matters, this implementation handles both strict and
; non-strict relations. It does this by also having a minimum and maximum 
; psuedo that a point can be equal to - a range of points that might be
; equal to this one. To tell the relation between any two points in the
; same chain, the psuedotime of the points is compared to see which is first,
; and then if the psuedotime of one point falls in the minimum-maximum range
; of the other, the relation is nonstrict; otherwise it is strict.
; When testing ranges, the end points are not included. So if the range
; is from 2000 to 4500, 2000 and 4500 are not considered to be in the range.
; There is also a "dummy" minimum and maximum corresponding to -infinity (nil)
; and +infinity (t).
; Minimum psuedo's are propagated forward, and maximum psuedo's backward, 
; similar to the propagation of absolute time and duration.
;
; The actual implementation points to the minimum or maximum point itself
; rather than just storing the psuedo times - this makes renumbering much
; easier.
;
; Example:
;  point           pt1      <=      pt2     <     pt3
;  psuedo           1               1000          2000
;  min-psuedo      nil               nil          1000
;  max-psuedo      2000             2000           t
;
; pt1 <= pt2 since 1000 is in range nil,2000, and 1 < 1000
; pt2 < pt3 since 2000 is not in range nil,2000, and 1000 < 2000
; similarly pt1 < pt3

;           Psuedo Time Calculations

; Psuedo-before calculates a new psuedo time before the given pt

(defun psuedo-before (pt)
  (if (equal (time-node-psuedo pt) newpsuedo)
      (- '0 psuedo-increment)
      (- (time-node-psuedo pt) psuedo-increment)
      )
  )

; Psuedo-after calculates a new psuedo time after the given pt

(defun psuedo-after (pt)
  (if (equal (time-node-psuedo pt) newpsuedo)
      psuedo-increment
      (+ (time-node-psuedo pt) psuedo-increment)
      )
  )

; Renumbering requires only that first descendant be used, as they are
; ordered, and if there is more than one, it means there are transitive edges 
; and they will be handled later anyway
 
; Renumber-next calculates the new psuedo time for the next point, and recurses

(defun renumber-next (last dlist)
  (unless (endp dlist)
          (let* ((p (time-point (time-link-to-point-name (car dlist)))))
	    (setf (time-node-psuedo p) (psuedo-after last))
	    (renumber-next p (time-node-descendants p))
	    )
	  )
  )

; Renumber renumbers  the psuedo times in a chain 

(defun renumber (meta &aux
                      (f (meta-node-first meta)))
  (setf (time-node-psuedo f) newpsuedo)
  (renumber-next f (time-node-descendants f))
  )

; Psuedo-between calculates a psuedo time between pt1 and pt2 using 90% of the
; difference. If there isn't enough room left between, the chain is renumbered.

(defun psuedo-between (pt1 pt2 &aux
                          (p1 (time-node-psuedo pt1))
                          (p2 (time-node-psuedo pt2)))
  (when (< (abs (- p2 p1)) '10)
        (renumber (time-node-chain pt1))
        (setq p1 (time-node-psuedo pt1))
        (setq p2 (time-node-psuedo pt2))
        )
  (when (equal p1 newpsuedo) (setq p1 '0))
  (when (equal p2 newpsuedo) (setq p2 '0))
  (+ (floor (/ (* (- p2 p1) '9) '10)) p1)
  )

; P-after returns t if x is greater than start (psuedo times)

(defun p-after (x start)
  (if (null start)
      t  
      (> x start)
      )  
  )

; P-before returns t if x is less than start (psuedo times)

(defun p-before (x end)
  (if (equal end t)
      t  
      (< x end)
      )  
  )

; Min-psuedo returns the minimum psuedo that this point can be equal to
; if the minimum point is -infinity (nil) return that, otherwise look up
; the psuedo time of the minimum point itself

(defun min-psuedo (pt &aux
                      (minptname (time-node-min-point-name pt)))
  (if (null minptname)
      nil
      (time-node-psuedo (time-point minptname))
      )
  )

; Max-psuedo returns the maximum psuedo that this point can be equal to
; if the maximum point is +infinity (nil) return that, otherwise look up
; the psuedo time of the maximum point itself

(defun max-psuedo (pt &aux
                      (maxptname (time-node-max-point-name pt)))
  (if (equal maxptname t)
      t  
      (time-node-psuedo (time-point maxptname))
      )  
  )

; Possibly-equal returns t if t1 and t2 can possibly be equal -i.e. they
; are <= or >=. The test is done by checking to see if the psuedo time of
; t2 fits in the range of psuedos defined by the min and max psuedos of t1
 
(defun possibly-equal (t1 t2 &aux
                          (p2 (time-node-psuedo t2)))
  (and (p-after p2 (min-psuedo t1))
       (p-before p2 (max-psuedo t1)))
  )

